package pwc.taxtech.atms.service.impl;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.StringUtils;

import pwc.taxtech.atms.exception.ApplicationException;
import pwc.taxtech.atms.common.AuthUserHelper;
import pwc.taxtech.atms.common.CommonConstants;
import pwc.taxtech.atms.common.CommonUtils;
import pwc.taxtech.atms.common.OperateLogType;
import pwc.taxtech.atms.common.OperationAction;
import pwc.taxtech.atms.common.OperationModule;
import pwc.taxtech.atms.dao.BusinessUnitMapper;
import pwc.taxtech.atms.dto.BusinessUnitDto;
import pwc.taxtech.atms.dto.BusinessUnitInputDto;
import pwc.taxtech.atms.dto.IdModel;
import pwc.taxtech.atms.dto.OperationLogDto;
import pwc.taxtech.atms.dto.UpdateLogParams;
import pwc.taxtech.atms.entity.BusinessUnit;
import pwc.taxtech.atms.entity.BusinessUnitExample;
import pwc.taxtech.atms.entity.BusinessUnitExample.Criteria;
import pwc.taxtech.atms.service.BusinessUnitService;
import pwc.taxtech.atms.service.OperationLogService;
import pwc.taxtech.atms.service.OrganizationService;

/**
 */
@Service
public class BusinessUnitServiceImpl implements BusinessUnitService {

    private static final Logger logger = LoggerFactory.getLogger(BusinessUnitServiceImpl.class);

    @Autowired
    private BusinessUnitMapper businessUnitMapper;

    @Autowired
    private OperationLogService operationLogService;

    @Autowired
    private OrganizationService organizationService;

    @Autowired
    private AuthUserHelper authUserHelper;

    @Override
    public List<BusinessUnitDto> getBusinessUnits() {
        logger.debug("BusinessUnitServiceImpl getBusinessUnits");
        BusinessUnitExample businessUnitExample = new BusinessUnitExample();
        businessUnitExample.setOrderByClause("name ASC");
        List<BusinessUnit> businessUnitList = businessUnitMapper.selectByExample(businessUnitExample);
        return rotateBusinessUnitList(businessUnitList);
    }

    @Transactional
    @Override
    public void addBusinessUnits(List<BusinessUnitInputDto> businessUnitDtoList) {
        logger.debug("BusinessUnitServiceImpl addBusinessUnits");
        if (businessUnitDtoList.isEmpty()) {
            throw new ApplicationException(CommonConstants.JSONNULLOBJECT);
        }
        for (BusinessUnitInputDto businessUnitDto : businessUnitDtoList) {
            addBusinessUnit(businessUnitDto);
        }
    }

    public void addBusinessUnit(BusinessUnitInputDto businessUnitDto) {
        BusinessUnit businessUnit = rotateBusinessUnitDtoForInput(businessUnitDto);
        businessUnit.setID(CommonUtils.getUUID());
        businessUnit.setCreateTime(new Date());
        businessUnit.setUpdateTime(new Date());
        businessUnitMapper.insert(businessUnit);

        // Log operation
        OperationLogDto operationLogDto = new OperationLogDto();
        operationLogDto.setOperationObject(businessUnitDto.getName());
        operationLogDto.setModule(OperationModule.BasicDataBusinessUnit.value());
        operationLogDto.setAction(OperationAction.New.value());
        operationLogDto.setLogType(OperateLogType.OperationLogBasicData.value());
        operationLogService.addOperationLog(operationLogDto);
    }

    @Transactional
    @Override
    public Integer deleteBusinessUnit(IdModel idModel) {
        logger.debug("BusinessUnitServiceImpl deleteBusinessUnit");

        if (idModel == null || !StringUtils.hasText(idModel.getID())) {
            throw new ApplicationException(CommonConstants.JSONNULLOBJECT);
        }

        BusinessUnit businessUnit = businessUnitMapper.selectByPrimaryKey(idModel.getID());

        if (businessUnit == null) {
            throw new ApplicationException("can't find business unit to delete, id: " + idModel.getID());
        }
        int result = businessUnitMapper.deleteByPrimaryKey(idModel.getID());

        // Log operation
        OperationLogDto operationLogDto = new OperationLogDto();
        operationLogDto.setModule(OperationModule.BasicDataBusinessUnit.value());
        operationLogDto.setAction(OperationAction.Delete.value());
        operationLogDto.setLogType(OperateLogType.OperationLogBasicData.value());
        operationLogDto.setOperationObject("");
        operationLogService.addOperationLog(operationLogDto);

        return result;
    }

    @Transactional
    @Override
    public Boolean updateBusinessUnits(List<BusinessUnitDto> businessUnitDtoList) {
        logger.debug("BusinessUnitServiceImpl updateBusinessUnits");

        if (businessUnitDtoList.isEmpty()) {
            throw new ApplicationException(CommonConstants.JSONNULLOBJECT);
        }

        for (BusinessUnitDto businessUnitDto : businessUnitDtoList) {
            // isActive changed or not
            boolean isStatusChangeOperation = false;

            if (businessUnitDto == null || !StringUtils.hasText(businessUnitDto.getID())) {
                throw new ApplicationException("primary key can't be null");
            }

            // current businessUnit
            BusinessUnit businessUnit = businessUnitMapper.selectByPrimaryKey(businessUnitDto.getID());
            if (businessUnit == null) {
                throw new ApplicationException("can't find business unit, id: " + businessUnitDto.getID());
            }

            // copy current businessUnit as tmp
            BusinessUnit originBusinessUnit = new BusinessUnit();
            CommonUtils.copyProperties(businessUnit, originBusinessUnit);

            if (businessUnitDto.getIsActive() != businessUnit.getIsActive()
                    || businessUnitDto.getName() != businessUnit.getName()) {
                isStatusChangeOperation = true;
                businessUnit.setIsActive(businessUnitDto.getIsActive());
                businessUnit.setName(businessUnitDto.getName());
                if (!businessUnitDto.getIsActive()
                        && organizationService.isOrganizationStructureExists(businessUnitDto.getID())) {
                    continue;
                }
            }
            businessUnit.setUpdateTime(new Date());
            businessUnitMapper.updateByPrimaryKey(businessUnit);

            if (isStatusChangeOperation) {
                UpdateLogParams updateLogParams = new UpdateLogParams();
                updateLogParams.setOperationObject(originBusinessUnit.getName());
                updateLogParams.setOriginalState(originBusinessUnit);
                updateLogParams.setUpdateState(businessUnit);
                updateLogParams.setOperationUser(authUserHelper.getCurrentAuditor());
                updateLogParams.setOperationModule(OperationModule.BasicDataBusinessUnit.value());
                updateLogParams.setOperationAction(OperationAction.Update.value());
                updateLogParams.setOperateLogType(OperateLogType.OperationLogBasicData.value());
                // check how many properties being updated and the log number should be same
                operationLogService.updateDataAddLog(updateLogParams);
            }
        }

        return true;
    }

    private List<BusinessUnitDto> rotateBusinessUnitList(List<BusinessUnit> businessUnitList) {
        List<BusinessUnitDto> businessUnitDtoList = new ArrayList<>();
        if (businessUnitList != null && !businessUnitList.isEmpty()) {
            for (BusinessUnit businessUnit : businessUnitList) {
                businessUnitDtoList.add(rotateBusinessUnit(businessUnit));
            }
        }
        return businessUnitDtoList;
    }

    private BusinessUnitDto rotateBusinessUnit(BusinessUnit businessUnit) {
        BusinessUnitDto businessUnitDto = new BusinessUnitDto();
        businessUnitDto.setID(businessUnit.getID());
        businessUnitDto.setIsActive(businessUnit.getIsActive());
        businessUnitDto.setName(businessUnit.getName());
        return businessUnitDto;
    }

    private BusinessUnit rotateBusinessUnitDto(BusinessUnitDto businessUnitDto) {
        BusinessUnit businessUnit = new BusinessUnit();
        businessUnit.setID(businessUnitDto.getID());
        businessUnit.setName(businessUnitDto.getName());
        businessUnit.setIsActive(businessUnitDto.getIsActive());
        return businessUnit;
    }
    
    private BusinessUnit rotateBusinessUnitDtoForInput(BusinessUnitInputDto businessUnitDto) {
        BusinessUnit businessUnit = new BusinessUnit();
        businessUnit.setID(businessUnitDto.getID());
        businessUnit.setName(businessUnitDto.getName());
        businessUnit.setIsActive(businessUnitDto.getIsActive());
        return businessUnit;
    }

    @Override
    public List<BusinessUnit> findByNameAndIsActive(String name, Boolean isActive) {
        logger.debug("This is a method similar to JPA which find business units by name and isActive");
        BusinessUnitExample example = new BusinessUnitExample();
        Criteria criteria = example.createCriteria();
        criteria.andNameEqualTo(name);
        criteria.andIsActiveEqualTo(isActive);

        return businessUnitMapper.selectByExample(example);
    }
    
    @Override
    public List<BusinessUnit> findByIsActiveTrue() {
        BusinessUnitExample example = new BusinessUnitExample();
        example.createCriteria().andIsActiveEqualTo(true);
        return businessUnitMapper.selectByExample(example);
    }

}
